"""
Tests for the Registry library.
"""
import asyncio
import unittest

from aioregistry.client import AsyncRegistryClient
from aioregistry.exceptions import RegistryException
from aioregistry.models import Descriptor, Registry
from aioregistry.parsing import parse_image_name


def asyncio_run(corofunc):
    """
    Simple decorator that starts an event loop to run a coroutine function.
    """

    def invoke(*args, **kwargs):
        return asyncio.run(corofunc(*args, **kwargs))

    return invoke


class RegistryTest(unittest.TestCase):
    """
    Tests for the Registry library.
    """

    def _check_image(self, name, registry, repo, ref="latest") -> None:
        """
        Test assertions an image name parse.
        """
        manifest_ref = parse_image_name(name)
        self.assertEqual(manifest_ref.registry, registry)
        self.assertEqual(manifest_ref.repo, repo)
        self.assertEqual(manifest_ref.ref, ref)

    def test_parse_image_ref_handling(self) -> None:
        """Test handling of refs"""
        # Test basic ref handling
        self._check_image(
            "ubuntu",
            None,
            ["library", "ubuntu"],
        )
        self._check_image(
            "docker.io/library/ubuntu",
            Registry(
                host="registry-1.docker.io",
                port=443,
                prot="https",
                host_alias="docker.io",
            ),
            ["library", "ubuntu"],
        )
        self._check_image(
            "ubuntu:18.04",
            None,
            ["library", "ubuntu"],
            ref="18.04",
        )
        self._check_image(
            "ubuntu@sha256:9b1702dcfe32c873a770a32cfd306dd7fc1c4fd134adfb783db68defc8894b3c",
            None,
            ["library", "ubuntu"],
            ref="sha256:9b1702dcfe32c873a770a32cfd306dd7fc1c4fd134adfb783db68defc8894b3c",
        )
        self._check_image(
            "cbir.clinc.ai/clinc/worker/gpu:v0.7.9",
            Registry(host="cbir.clinc.ai"),
            ["clinc", "worker", "gpu"],
            ref="v0.7.9",
        )

    def test_parse_image_localhost_handling(self) -> None:
        """Test handling of localhost hosted image names"""
        # localhost is special
        self._check_image(
            "localhost/msg",
            Registry(host="localhost", port=80, prot="http"),
            ["msg"],
        )
        self._check_image(
            "notahost/msg",
            None,
            ["notahost", "msg"],
        )
        self._check_image(
            "isa.host/msg",
            Registry(host="isa.host"),
            ["msg"],
        )

        # Test localhost/isahost with different explicit ports.
        self._check_image(
            "localhost:555/msg",
            Registry(
                host="localhost", port=555, prot="http", host_alias="localhost:555"
            ),
            ["msg"],
        )
        self._check_image(
            "isahost:555/msg",
            Registry(host="isahost", port=555, prot="https", host_alias="isahost:555"),
            ["msg"],
        )
        self._check_image(
            "localhost:443/msg",
            Registry(
                host="localhost", port=443, prot="https", host_alias="localhost:443"
            ),
            ["msg"],
        )
        self._check_image(
            "isahost:443/msg",
            Registry(host="isahost", port=443, prot="https", host_alias="isahost:443"),
            ["msg"],
        )
        self._check_image(
            "localhost:80/msg",
            Registry(host="localhost", port=80, prot="http", host_alias="localhost:80"),
            ["msg"],
        )
        self._check_image(
            "isahost:80/msg",
            Registry(host="isahost", port=80, prot="http", host_alias="isahost:80"),
            ["msg"],
        )

    def test_parse_image_explicit_protocol(self) -> None:
        """Test handling of localhost hosted image names"""
        # Explicit protocol
        self._check_image(
            "http://localhost/msg",
            Registry(
                host="localhost", port=80, prot="http", host_alias="http://localhost"
            ),
            ["msg"],
        )
        self._check_image(
            "http://isahost/msg",
            Registry(host="isahost", port=80, prot="http", host_alias="http://isahost"),
            ["msg"],
        )
        self._check_image(
            "https://localhost/msg",
            Registry(
                host="localhost", port=443, prot="https", host_alias="https://localhost"
            ),
            ["msg"],
        )
        self._check_image(
            "https://isahost/msg",
            Registry(
                host="isahost", port=443, prot="https", host_alias="https://isahost"
            ),
            ["msg"],
        )

        self._check_image(
            "myregistry:555/msg",
            Registry(
                host="myregistry", port=555, prot="https", host_alias="myregistry:555"
            ),
            ["msg"],
        )
        self._check_image(
            "localhost:443/msg",
            Registry(
                host="localhost", port=443, prot="https", host_alias="localhost:443"
            ),
            ["msg"],
        )

    def test_registry_str(self) -> None:
        """
        Test Registry's string representation
        """
        self.assertEqual(str(Registry(host="abc")), "abc")
        self.assertEqual(str(Registry(host="abc-123", host_alias="abc")), "abc")
        self.assertEqual(str(Registry(host="abc", port=80, prot="http")), "abc:80")
        self.assertEqual(
            str(Registry(host="abc", port=80, prot="https")), "https://abc:80"
        )
        self.assertEqual(
            str(Registry(host="abc", port=443, prot="http")), "http://abc:443"
        )
        self.assertEqual(str(Registry(host="abc", port=443, prot="https")), "abc")
        self.assertEqual(
            str(Registry(host="abc", port=555, prot="http")), "http://abc:555"
        )
        self.assertEqual(str(Registry(host="abc", port=555, prot="https")), "abc:555")

        self.assertEqual(str(Registry(host="localhost")), "localhost:443")
        self.assertEqual(
            str(Registry(host="localhost", port=80, prot="http")), "localhost"
        )
        self.assertEqual(
            str(Registry(host="localhost", port=80, prot="https")),
            "https://localhost:80",
        )
        self.assertEqual(
            str(Registry(host="localhost", port=443, prot="http")),
            "http://localhost:443",
        )
        self.assertEqual(
            str(Registry(host="localhost", port=443, prot="https")), "localhost:443"
        )
        self.assertEqual(
            str(Registry(host="localhost", port=555, prot="http")), "localhost:555"
        )
        self.assertEqual(
            str(Registry(host="localhost", port=555, prot="https")),
            "https://localhost:555",
        )

    @asyncio_run
    async def test_exists(self) -> None:
        """Test behavior of ManifestRef.exists()"""
        async with AsyncRegistryClient() as client:
            self.assertEqual(
                await client.ref_lookup(
                    parse_image_name("msg555/aioregistry-test:test_tag")
                ),
                Descriptor(
                    mediaType="application/vnd.docker.distribution.manifest.v2+json",
                    digest="sha256:359966c4859d1b2cbff80120cf25e0d1dfcd0c03efa398754da4af9a5b9fa255",
                    size=524,
                ),
            )
            self.assertIsNone(
                await client.ref_lookup(
                    parse_image_name("msg555/aioregistry-test:fake_tag")
                )
            )
            with self.assertRaises(RegistryException):
                await client.ref_lookup(
                    parse_image_name("fake.repo/msg555/aioregistry-test:fake_tag")
                )


if __name__ == "__main__":
    unittest.main()



import asyncio
from typing import Any


async def async_generator_buffer(gen, buffer_size):
    """
    Automatically buffer a few elements of an async generator and yield
    the buffered results as another async generator.
    """

    cond = asyncio.Condition()

    output_pos = 0
    output_size = 0
    output_complete = False
    outputs = [None for _ in range(buffer_size)]

    async def _fill_buffer():
        nonlocal output_pos, output_size, output_complete
        try:
            async for value in gen:
                async with cond:
                    while output_size == buffer_size:
                        await cond.wait()
                    outputs[(output_pos + output_size) % buffer_size] = value
                    output_size += 1
                    cond.notify()
        finally:
            async with cond:
                output_complete = True
                cond.notify()

    fill_buffer_task = asyncio.create_task(_fill_buffer())

    while True:
        async with cond:
            while not output_complete and output_size == 0:
                await cond.wait()

            if output_complete and output_size == 0:
                break

            value = outputs[output_pos]
            output_pos = (output_pos + 1) % buffer_size
            output_size -= 1
            cond.notify()
        yield value

    # Make sure task has completed, propagate any failure
    await fill_buffer_task


class ReleaseableAsyncContextManager:
    """
    A context manager that allows for a context to be released and then
    re-entered. This is useful when you want to explicitly return a context
    manager that you already entered back to the client and leave them
    responsible for closing the context.
    """

    NO_ENTRY_VALUE = object()

    def __init__(
        self,
        acm,
        *,
        value: Any = NO_ENTRY_VALUE,
    ) -> None:
        self.acm = acm
        self.value = value

    async def __aenter__(self):
        if self.value is self.NO_ENTRY_VALUE:
            self.value = await self.acm.__aenter__()
        return self.value

    async def __aexit__(self, exc_typ, exc_val, exc_tb) -> None:
        if self.acm is not None:
            await self.acm.__aexit__(exc_typ, exc_val, exc_tb)

    def release(self):
        """
        Create a new context manager that wraps the currently wrapped
        context manager in a way that the caller is now responsible for
        releasing the resources of the wrapped context manager.
        """
        acm = self.acm
        self.acm = None
        return ReleaseableAsyncContextManager(acm, value=self.value)



import hashlib
import re
from typing import Any, ClassVar, Dict, List, Literal, Optional, Tuple

from pydantic import BaseModel, Field

# https://docs.docker.com/registry/spec/manifest-v2-2/
# https://github.com/opencontainers/image-spec/blob/main/media-types.md#compatibility-matrix
MANIFEST_TYPE_MAP = {}


class Descriptor(BaseModel, frozen=True):
    """
    Generic descriptor model used throughout the manifest definitions. These
    objects point to a content addressed object stored elsewhere.
    """

    media_type: str = Field(..., alias="mediaType")
    size: int
    digest: str
    urls: List[str] = []
    annotations: Dict[str, str] = {}


class Manifest(BaseModel, frozen=True):
    """
    Base Manifest class that supplies some useful methods.
    """

    MEDIA_TYPES: ClassVar[Tuple[str, ...]]

    @classmethod
    def __init_subclass__(cls, *args, **kwargs) -> None:
        for media_type in getattr(cls, "MEDIA_TYPES", ()):
            MANIFEST_TYPE_MAP[media_type] = cls
        super().__init_subclass__(*args, **kwargs)

    @classmethod
    def parse(
        cls,
        data: Any,
        *,
        media_type: Optional[str] = None,
    ) -> "Manifest":
        """
        Attempt to parse data as a manifest object. If the media type is
        known it can be specified to only try the specified media type.
        """
        if media_type is None:
            if not isinstance(data, dict):
                raise ValueError("data is not a dict")
            media_type = data.get("mediaType")
            if media_type is None:
                raise ValueError("data has no media type and none given")

        manifest_cls = MANIFEST_TYPE_MAP.get(media_type)
        if manifest_cls is None:
            raise ValueError(f"Unknown media type {repr(media_type)}")
        return manifest_cls(**data)

    @property
    def digest(self) -> str:
        """
        Compute the manifest digest using its canonicalized form. This may
        differ from the digest used on the registry if the server was using
        a different canonicalization (which at this point seems likely).
        """
        digest = self.__dict__.get("_digest")
        if digest is not None:
            return digest

        h = hashlib.sha256()
        h.update(self.canonical().encode("utf-8"))
        digest = "sha256:" + h.hexdigest()
        self.__dict__["_digest"] = digest
        return digest

    def canonical(self) -> str:
        """
        Calculate the canonical JSON representation.
        """
        if self.get_media_type().startswith("application/vnd.docker."):
            return self.json(
                exclude_unset=True,
                indent=3,
                separators=(",", ": "),
                ensure_ascii=False,
                by_alias=True,
            )
        return self.json(
            exclude_unset=True,
            ensure_ascii=False,
            by_alias=True,
        )

    def get_media_type(self) -> str:
        """
        Returns the media type of the manifest. Most manifest types have a
        media_type member that this is fulfilled with but V1 schema manifests
        lack this field which this method compensates for.
        """
        # pylint: disable=no-member
        if isinstance(self, ManifestV1):
            return self.MEDIA_TYPES[0]
        return self.media_type  # type: ignore

    def get_manifest_dependencies(self) -> List[str]:
        """Return a list of manifest dependency digests."""
        return []

    def get_blob_dependencies(self) -> List[str]:
        """Return a list of blob dependency digests."""
        return []


class ManifestListV2S2(Manifest, frozen=True):
    """
    Manifest list type
    """

    MEDIA_TYPES = (
        "application/vnd.oci.image.index.v1+json",
        "application/vnd.docker.distribution.manifest.list.v2+json",
    )

    class ManifestListItem(Descriptor, frozen=True):
        """Container class for a sub-manifest in a manifest list"""

        class PlatformData(BaseModel, frozen=True):
            """Container class for platform data in a manifest list"""

            architecture: str
            os: str
            os_version: str = Field("", alias="os.version")
            os_features: List[str] = Field([], alias="os.features")
            variant: str = ""
            features: List[str] = []

        platform: Optional[PlatformData] = None

    schema_version: Literal[2] = Field(..., alias="schemaVersion")
    media_type: Optional[Literal[MEDIA_TYPES]] = Field(  # type: ignore
        MEDIA_TYPES[0], alias="mediaType"
    )
    manifests: List[ManifestListItem]
    annotations: Dict[str, str] = {}

    def get_manifest_dependencies(self) -> List[str]:
        """Return a list of manifest dependency digests."""
        return [manifest.digest for manifest in self.manifests]


class ManifestV2S2(Manifest, frozen=True):
    """
    Single image manifest
    """

    MEDIA_TYPES = (
        "application/vnd.oci.image.manifest.v1+json",
        "application/vnd.docker.distribution.manifest.v2+json",
    )

    schema_version: Literal[2] = Field(..., alias="schemaVersion")
    media_type: Literal[MEDIA_TYPES] = Field(MEDIA_TYPES[0], alias="mediaType")  # type: ignore
    config: Descriptor
    layers: List[Descriptor]

    def get_blob_dependencies(self) -> List[str]:
        """Return a list of manifest dependency digests."""
        result = [layer.digest for layer in self.layers]
        result.append(self.config.digest)
        return result


class ManifestV1(Manifest, frozen=True):
    """
    Legacy manifest type.

    Although we can accept signed V1 manifests there is no support for verifiying
    the signatures attached. Currently the signatures are just dropped. Since this
    is a legacy media type support is unlikely to be added.
    """

    MEDIA_TYPES = (
        "application/vnd.docker.distribution.manifest.v1+json",
        "application/vnd.docker.distribution.manifest.v1+prettyjws",
    )

    class BlobData(BaseModel, frozen=True):
        """Container class for manifest blob data"""

        blob_sum: str = Field(..., alias="blobSum")

    class HistoryData(BaseModel, frozen=True):
        """Container class for manifest history data"""

        v1_compatibility: str = Field(..., alias="v1Compatibility")

    name: str
    tag: str
    architecture: str
    fsLayers: List[BlobData]
    history: List[HistoryData]
    schemaVersion: int


class Registry(BaseModel, frozen=True):
    """
    Represents a docker registry.
    """

    host: str
    port: int = 443
    prot: str = "https"
    host_alias: Optional[str] = None

    def __str__(self) -> str:
        if self.host_alias is not None:
            return self.host_alias
        default_prot, default_port = "https", 443
        if self.host == "localhost":
            default_prot, default_port = "http", 80
        if self.prot == default_prot and self.port == default_port:
            return self.host
        if self.prot == default_prot and self.port not in (80, 443):
            return f"{self.host}:{self.port}"
        if self.prot != default_prot and default_port + self.port == 80 + 443:
            return f"{self.host}:{self.port}"
        return f"{self.prot}://{self.host}:{self.port}"

    @property
    def url(self) -> str:
        """
        Returns the base url of the registry.
        """
        return f"{self.prot}://{self.host}:{self.port}"


class RegistryBlobRef(BaseModel, frozen=True):
    """
    Represents a blob ref on a registry.
    """

    OBJECT_TYPE: ClassVar[str] = "blobs"

    registry: Optional[Registry]
    repo: List[str]
    ref: str

    @property
    def url(self) -> str:
        """
        Returns the path component of the blob url underneath the registry.
        """
        return f"v2/{'/'.join(self.repo)}/{self.OBJECT_TYPE}/{self.ref}"

    def upload_url(self, upload_uuid: str = "") -> str:
        """
        Returns the url path that should be used to initiate a blob upload.
        """
        return f"v2/{'/'.join(self.repo)}/{self.OBJECT_TYPE}/uploads/{upload_uuid}"

    def is_digest_ref(self) -> bool:
        """
        Returns true if ref is a disgest ref.
        """
        return bool(re.fullmatch(r"sha256:[0-9a-f]{64}", self.ref))

    def __str__(self) -> str:
        return self.name(truncate=True)

    def name(self, *, truncate=False, include_ref=True) -> str:
        """Return the full blob name"""
        repo_name = "/".join(self.repo)
        if self.registry:
            repo_name = f"{self.registry}/{repo_name}"
        if not include_ref:
            return repo_name
        if self.is_digest_ref():
            return f"{repo_name}@{self.ref[7:14] if truncate else self.ref}"
        return f"{repo_name}:{self.ref}"


class RegistryManifestRef(RegistryBlobRef, frozen=True):
    """
    Represents a manifest ref in a registry.
    """

    OBJECT_TYPE = "manifests"



"""
Tests for the aioregistry.auth module
"""
import asyncio
import base64
import unittest
from unittest.mock import patch

from aioregistry.auth import DictCredentialStore, DockerCredentialStore


def asyncio_run(corofunc):
    """
    Simple decorator that starts an event loop to run a coroutine function.
    """

    def invoke(*args, **kwargs):
        return asyncio.run(corofunc(*args, **kwargs))

    return invoke


class AuthTest(unittest.TestCase):
    """
    aioregistry.auth tests
    """

    @asyncio_run
    async def test_dict_store(self):
        """
        Test that dict cred store does yields the configured auths.
        """
        creds = {
            "test.io": ("myuser", "mypass"),
            "docker.io": ("myuser1", "mypass2"),
        }
        cred_store = DictCredentialStore(creds)

        for host, auth in creds.items():
            self.assertEqual(
                await cred_store.get(host),
                auth,
            )

        self.assertIsNone(await cred_store.get("other.io"))

    @asyncio_run
    async def test_docker_auth_map(self):
        """
        Test that basic docker auth config works.
        """
        creds = {
            "test.io": ("myuser", "mypass"),
            "https://index.docker.io/v1/": ("myuser1", "mypass2"),
        }
        # pylint wtf?
        # pylint: disable=redefined-variable-type
        cred_store = DockerCredentialStore(
            {
                "auths": {
                    host: {"auth": base64.b64encode(":".join(auth).encode()).decode()}
                    for host, auth in creds.items()
                }
            }
        )

        for host, auth in creds.items():
            self.assertEqual(
                await cred_store.get(host),
                auth,
            )

        self.assertIsNone(await cred_store.get("other.io"))

    @asyncio_run
    async def test_docker_cred_helper(self):
        """
        Test using credential helper configurations.
        """
        cred_store = DockerCredentialStore(
            {
                "auths": {"test.io": {}},
                "credsStore": "test-store",
                "credHelpers": {"reg1": "helper1", "reg2": "helper2"},
            }
        )

        async def _test_query(
            cred_helper, host, user, password, expect_cached=False, query_host=None
        ):
            query_host = query_host or host
            with patch("aioregistry.auth.DockerCredentialStore._query_helper") as qhelp:
                qhelp.return_value = {"Username": user, "Secret": password}
                self.assertEqual(
                    await cred_store.get(host),
                    (user, password) if user else None,
                )
                if expect_cached:
                    qhelp.assert_not_called()
                else:
                    qhelp.assert_called_once_with(cred_helper, query_host)

        await _test_query("test-store", "test.io", "user1", "pass1")
        await _test_query("test-store", "test.io", "user1", "pass1", expect_cached=True)
        await _test_query(
            "test-store",
            "docker.io",
            "user2",
            "pass2",
            query_host="https://index.docker.io/v1/",
        )
        await _test_query("helper1", "reg1", "user3", "pass3")
        await _test_query("helper2", "reg2", "", "")
        await _test_query("helper2", "reg2", "", "", expect_cached=True)



"""
Defines the CredentialStore interface responsible for providing authentication
to registries.
"""
import abc
import asyncio
import base64
import json
import logging
import os
import subprocess
from datetime import datetime, timedelta
from typing import Any, Awaitable, Dict, Mapping, Optional, Tuple

LOGGER = logging.getLogger(__name__)


class CredentialStore(metaclass=abc.ABCMeta):
    """
    Interface for accessing a registry credential store.
    """

    @abc.abstractmethod
    async def get(self, host: str) -> Optional[Tuple[str, str]]:
        """
        If credentials for the given host are available returns a
        (user, password) tuple containing those credentials. Otherwise
        returns None if no credentials exist for the requested host.

        If the credentials represent an identity token the `user` part of
        the tuple will be "<token>".
        """


class DockerCredentialStore(CredentialStore):
    """
    Implements a credential store that understands the containers auth
    config format. See
    https://github.com/containers/image/blob/main/docs/containers-auth.json.5.md
    for more information on this format.
    """

    HOST_REMAP = {
        "docker.io": "https://index.docker.io/v1/",
    }

    def __init__(self, config) -> None:
        self.default_store = config.get("credsStore")
        self.host_stores = config.get("credHelpers", {})
        self.auths = {
            host: tuple(base64.b64decode(auth["auth"]).decode().split(":", 1))
            for host, auth in config.get("auths", {}).items()
            if auth.get("auth")
        }

    @classmethod
    def from_file(cls, path: str) -> "DockerCredentialStore":
        """
        Load the docker config from a file.

        Raises FileNotFoundError if the file could not be opened.
        """
        with open(path, "r", encoding="utf-8") as fconfig:
            return cls(json.load(fconfig))

    @staticmethod
    async def _query_helper(store: str, host: str) -> Any:
        """
        Query the passed storage helper and return the parsed JSON results.

        If the invocation fails or the resutls are not a valid JSON document
        None will be returned instead.
        """
        LOGGER.info("Querying %s for %s", store, host)

        # Query the storage helper.
        try:
            presult = await asyncio.create_subprocess_exec(
                f"docker-credential-{store}",
                "get",
                stdin=subprocess.PIPE,
                stdout=subprocess.PIPE,
                stderr=subprocess.DEVNULL,
            )
            stdout, _ = await presult.communicate(host.encode("utf-8"))
            if presult.returncode != 0:
                return None
        except OSError:
            return None
        try:
            return json.loads(stdout.decode("utf-8"))
        except ValueError:
            return None

    async def get(self, host: str) -> Optional[Tuple[str, str]]:
        """
        Gets the credentials for the host. First it checks if the credentials
        are stored literally or have been cached in self.auths. Next we check if
        there is a host specific credential store configured, otherwise we default
        to the global credential store helper. If either of those things exist
        we'll attempt to query the helper and cache the result if the program
        exits correctly.
        """
        # If there is a remap for the host try that first, then try the host
        # itself.
        remap_host = self.HOST_REMAP.get(host)
        if remap_host is not None:
            if remap_result := await self.get(remap_host):
                return remap_result

        # Return credentials from cache if they exist.
        auth = self.auths.get(host, False)
        if auth is not False:
            return auth  # type: ignore

        # Determine what storage helper we should use.
        store = self.host_stores.get(host, self.default_store)
        if store is None:
            return None

        # Query the storage helper.
        result = await self._query_helper(store, host)
        if result is None:
            return None

        username = result.get("Username")
        password = result.get("Secret")

        # Cache and return results.
        result = None
        if username and password:
            result = (username, password)
        self.auths[host] = result
        return result


class DictCredentialStore(CredentialStore):
    """
    Simple in-memory dictionary backed credential store.
    """

    def __init__(self, auth_map: Mapping[str, Tuple[str, str]]) -> None:
        self.auth_map = dict(auth_map)

    async def get(self, host: str) -> Optional[Tuple[str, str]]:
        """
        Return the credentials in the dict if they exist.
        """
        return self.auth_map.get(host)


class ChainedCredentialStore(CredentialStore):
    """
    Chains multiple other credential stores together. The earliest providers
    will be queried first. If credentials are successfully returned by a
    provider no other providers in the chain will be queried.
    """

    def __init__(self, *providers: CredentialStore) -> None:
        self.providers = list(providers)

    async def get(self, host: str) -> Optional[Tuple[str, str]]:
        """
        Find the first provider with credentials for the host.
        """
        for provider in self.providers:
            creds = await provider.get(host)
            if creds is not None:
                return creds
        return None


class CachingStore(CredentialStore):
    """
    Wrapper around a credential store that caches results for a TTL.
    """

    def __init__(self, provider: CredentialStore, ttl: timedelta) -> None:
        self.provider = provider
        self.ttl = ttl
        self.cache_data: Dict[str, Tuple[Awaitable[Tuple[str, str]], datetime]] = {}

    async def get(self, host: str) -> Optional[Tuple[str, str]]:
        """
        Check cache for result, go to provider if not cached.
        """
        now = datetime.now()
        creds_task, cache_time = self.cache_data.get(host, (None, None))
        if creds_task is None or cache_time is None or now - cache_time > self.ttl:
            creds_task = asyncio.create_task(self.provider.get(host))  # type: ignore
            self.cache_data[host] = (creds_task, now)

        return await creds_task


def default_credential_store(ttl=timedelta(minutes=1)) -> CredentialStore:
    """
    Creates a credential stored that queries all the paths described in
    https://github.com/containers/image/blob/main/docs/containers-auth.json.5.md
    """
    paths = []
    if runtime_dir := os.getenv("XDG_RUNTIME_DIR", ""):
        paths.append(os.path.join(runtime_dir, "containers/auth.json"))
    if config_home := os.getenv("XDG_CONFIG_HOME", ""):
        paths.append(os.path.join(config_home, ".config/containers/auth.json"))
    elif home_dir := os.getenv("HOME", ""):
        paths.append(os.path.join(home_dir, ".config/containers/auth.json"))
    if home_dir := os.getenv("HOME", ""):
        paths.append(os.path.join(home_dir, ".docker/config.json"))
        paths.append(os.path.join(home_dir, ".dockercfg"))

    providers = []
    for path in paths:
        try:
            with open(path, "r", encoding="utf-8") as fconfig:
                providers.append(DockerCredentialStore(json.load(fconfig)))
        except FileNotFoundError:
            pass
    return CachingStore(ChainedCredentialStore(*providers), ttl=ttl)



"""
Expose public aioregistry interface
"""
from .auth import (
    ChainedCredentialStore,
    CredentialStore,
    DictCredentialStore,
    DockerCredentialStore,
    default_credential_store,
)
from .client import AsyncRegistryClient
from .exceptions import RegistryException
from .models import (
    Descriptor,
    Manifest,
    ManifestListV2S2,
    ManifestV1,
    ManifestV2S2,
    Registry,
    RegistryBlobRef,
    RegistryManifestRef,
)
from .parsing import parse_image_name



"""
Module implementing a client for the v2 docker registry API.

See https://docs.docker.com/registry/spec/api/
"""
import asyncio
import hashlib
import json
import logging
import ssl
import urllib.parse
from typing import (
    Any,
    AsyncIterable,
    Awaitable,
    Callable,
    Dict,
    List,
    Optional,
    Tuple,
    Union,
)

import aiohttp
from pydantic import BaseModel, Field, ValidationError

from .auth import CredentialStore, DictCredentialStore
from .exceptions import RegistryException
from .models import (
    MANIFEST_TYPE_MAP,
    Descriptor,
    Manifest,
    Registry,
    RegistryBlobRef,
    RegistryManifestRef,
)
from .parsing import split_quote
from .utils import ReleaseableAsyncContextManager, async_generator_buffer

LOGGER = logging.getLogger(__name__)


class AccessTokenResponse(BaseModel):
    """
    Model to capture auth response
    """

    access_token: str = Field(..., alias="token")


def _get_descriptor_from_response(
    ref: RegistryBlobRef, response: aiohttp.ClientResponse
) -> Optional[Descriptor]:
    """
    Extract a descriptor from a response. Return None if the descriptor does
    not exist or if we are unauthorized to access it.
    """
    if response.status in (401, 404):
        return None
    if response.status != 200:
        raise RegistryException("Unexpected response from registry")

    # Extract digest
    digest: Optional[str]
    if ref.is_digest_ref():
        digest = ref.ref
    else:
        digest = response.headers.get("Docker-Content-Digest")
        if digest is None:
            raise RegistryException("No digest given by server for tag ref")

    # Extract media type
    media_type = response.headers.get("Content-Type")
    if media_type is None:
        raise RegistryException("No content type given by server")

    # Extract size
    size = response.headers.get("Content-Length")
    if size is None:
        raise RegistryException("No content length given by server")
    try:
        isize = int(size)
    except ValueError as exc:
        raise RegistryException("Invalid content length given by server") from exc

    return Descriptor(
        mediaType=media_type,
        size=isize,
        digest=digest,
    )


class AsyncRegistryClient:
    """
    Class that holds network session and context information.
    """

    _ACCEPT_HEADER = ",".join(MANIFEST_TYPE_MAP) + ", */*"
    _DEFAULT_REGISTRY = Registry(
        host="registry-1.docker.io",
        host_alias="docker.io",
    )
    _DEFAULT_TIMEOUT = aiohttp.ClientTimeout(
        total=None,
        connect=None,
        sock_connect=None,
        sock_read=None,
    )

    def __init__(
        self,
        *,
        session: Optional[aiohttp.ClientSession] = None,
        creds: Optional[CredentialStore] = None,
        timeout: Optional[aiohttp.ClientTimeout] = None,
        default_registry: Optional[Registry] = None,
        ssl_context: Optional[ssl.SSLContext] = None,
    ) -> None:
        self.custom_session = bool(session)
        self.session = session or aiohttp.ClientSession()
        self.timeout = timeout or self._DEFAULT_TIMEOUT
        self.default_registry = default_registry or self._DEFAULT_REGISTRY
        self.ssl_context = ssl_context
        self.access_tokens: Dict[Tuple[str, str], str] = {}
        self.creds = creds or DictCredentialStore({})

    async def __aenter__(self) -> "AsyncRegistryClient":
        return self

    async def __aexit__(self, exc_type, exc_value, exc_traceback) -> None:
        if not self.custom_session:
            await self.session.close()

    async def _request(
        self,
        method: str,
        registry: Registry,
        path: str,
        *,
        headers: Optional[Dict[str, str]] = None,
        data: Optional[Union[str, bytes]] = None,
    ):
        """
        Make a request to a registry, applying the appropriate credentials.

        Returns an async context manager that yields an aiohttp response.
        """
        # Parse URL and determine the the authentication key for any
        # authentication tokens.
        url = urllib.parse.urljoin(registry.url, path)
        url_data = urllib.parse.urlparse(url)
        path_parts = url_data.path.split("/")
        auth_key = (url_data.hostname or "", "/".join(path_parts[0:4]))

        # Lookup any basic auth credentials to supply.
        if not registry and url_data.hostname is None:
            raise ValueError("No registry or hostname provided")

        auth = None
        creds = await self.creds.get(
            registry.host_alias or registry.host if registry else url_data.hostname  # type: ignore
        )

        if creds is not None:
            auth = aiohttp.BasicAuth(creds[0], password=creds[1])

        # Attempt to make the request twice. If the first attempt fails with a
        # 401 try to get an authentication token and then try again.
        first_attempt = True
        while True:
            all_headers = dict(headers or {})
            all_headers["Accept"] = self._ACCEPT_HEADER

            basic_auth = None
            auth_token = self.access_tokens.get(auth_key)
            if auth_token is None:
                basic_auth = auth
            else:
                all_headers["Authorization"] = "Bearer " + auth_token

            acm = ReleaseableAsyncContextManager(
                self.session.request(
                    method,
                    url,
                    auth=basic_auth,
                    headers=all_headers,
                    data=data,
                    timeout=self.timeout,
                    ssl=self.ssl_context,
                )
            )
            async with acm as response:
                if not first_attempt or response.status != 401:
                    return acm.release()

            www_auth = response.headers.get("WWW-Authenticate", "")
            if not www_auth.startswith("Bearer "):
                raise RegistryException("Failed to make request, unauthorized")

            auth_parts = split_quote(www_auth[7:], "=,")
            auth_args = {
                auth_parts[i]: auth_parts[i + 2]
                for i in range(0, len(auth_parts) - 2, 4)
            }
            realm = auth_args.pop("realm", "")
            if realm is None:
                raise RegistryException("Expected authentication realm")

            query_parts = []
            if service := auth_args.get("service"):
                query_parts.append(("service", service))
            query_parts.extend(
                ("scope", scope)
                for scope in auth_args.get("scope", "").split(" ")
                if scope
            )

            async with self.session.get(
                realm + "?" + urllib.parse.urlencode(query_parts),
                auth=auth,
                timeout=self.timeout,
                ssl=self.ssl_context,
            ) as auth_resp:
                if auth_resp.status != 200:
                    raise RegistryException("Failed to generate authentication token")

                try:
                    auth_response_data = AccessTokenResponse.parse_raw(
                        await auth_resp.read()
                    )
                except ValidationError as exc:
                    raise RegistryException(
                        "Got unexpected auth data response"
                    ) from exc
                self.access_tokens[auth_key] = auth_response_data.access_token

            first_attempt = False

    async def ref_delete(self, ref: RegistryBlobRef) -> bool:
        """
        Attempts to delete a reference from the registry. This is only supported
        for non-digest manifest references.

        Returns True if the delete succeeded. If permission is denied or the ref
        does not exist returns False. Other failures will raise an exception.
        """
        registry = ref.registry or self.default_registry
        try:
            async with await self._request("DELETE", registry, ref.url) as response:
                if response.status in (401, 404):
                    return False
                if response.status != 202:
                    raise RegistryException("Delete failed")
                return True
        except aiohttp.ClientError as exc:
            raise RegistryException("failed to contact registry") from exc

    async def ref_lookup(self, ref: RegistryBlobRef) -> Optional[Descriptor]:
        """
        Attempts to lookup the requested ref and return a :class:`Descriptor`
        representation. The descriptor includes the media type, digest, and
        size information of the object.

        If the registry returns a 404 Not Found or 401 Unauthorized this method
        will return None. Any other failure to retrieve metadata about the
        object will raise an exception.
        """
        registry = ref.registry or self.default_registry
        try:
            async with await self._request("HEAD", registry, ref.url) as response:
                return _get_descriptor_from_response(ref, response)
        except aiohttp.ClientError as exc:
            raise RegistryException("failed to contact registry") from exc

    async def ref_content_stream(
        self,
        ref: RegistryBlobRef,
        chunk_size: int = 5 * 2**20,
    ) -> Tuple[Descriptor, AsyncIterable[bytes]]:
        """
        Returns a descriptor of the blob and an async iterable of the data that yields
        chunks of `chunk_size` bytes. The last chunk may be smaller than `chunk_size`.
        """

        async def _chunk_rest(acm: Any) -> AsyncIterable[bytes]:
            """Helper method to chunk the response data as an async iterable"""
            async with acm as response:
                cur_chunk: List[bytes] = []
                cur_chunk_size = 0
                async for chunk in response.content.iter_chunked(chunk_size):
                    need = chunk_size - cur_chunk_size
                    cur_chunk_size += len(chunk)
                    if len(chunk) >= need:
                        yield b"".join(cur_chunk) + chunk[:need]

                        cur_chunk.clear()
                        if need < len(chunk):
                            cur_chunk.append(chunk[need:])
                        cur_chunk_size -= chunk_size
                    else:
                        cur_chunk.append(chunk)

                if cur_chunk:
                    yield b"".join(cur_chunk)

        registry = ref.registry or self.default_registry
        try:
            acm = await self._request("GET", registry, ref.url)
            async with acm as response:
                descriptor = _get_descriptor_from_response(ref, response)
                if descriptor is None:
                    raise RegistryException("Could not find blob")
                return descriptor, _chunk_rest(acm.release())
        except aiohttp.ClientError as exc:
            raise RegistryException("failed to contact registry") from exc

    async def manifest_download(self, ref: RegistryManifestRef) -> Manifest:
        """
        Attempt to download a manifest.
        """
        registry = ref.registry or self.default_registry
        try:
            async with await self._request("GET", registry, ref.url) as response:
                if response.status != 200:
                    raise RegistryException(
                        f"Unexpected response from registry HTTP {response.status}"
                    )
                try:
                    manifest_data = json.loads(await response.text(encoding="utf-8"))
                except ValueError as exc:
                    raise RegistryException(
                        "Failed decoding JSON response from registry"
                    ) from exc
        except aiohttp.ClientError as exc:
            raise RegistryException("failed to contact registry") from exc

        return Manifest.parse(
            manifest_data,
            media_type=response.headers.get("Content-Type"),
        )

    async def manifest_write(
        self, ref: RegistryManifestRef, manifest: Manifest
    ) -> None:
        """
        Write a manifest to a registry. If `ref.ref` is empty or is a digest
        ref, then `ref.ref` will be ignored and the manifest will be pushed
        untagged using the digest of `manifest.canonical()`.
        """
        if not ref.ref or ref.is_digest_ref():
            ref = ref.copy(update={"ref": manifest.digest})
        async with await self._request(
            "PUT",
            ref.registry or self.default_registry,
            ref.url,
            data=manifest.canonical(),
            headers={"Content-Type": manifest.get_media_type()},
        ) as response:
            if response.status // 100 != 2:
                raise RegistryException("Failed to copy manifest")

    async def blob_write(
        self,
        ref: RegistryBlobRef,
        data: AsyncIterable[bytes],
        *,
        progress_callback: Optional[Callable[[int], Awaitable[None]]] = None,
    ) -> RegistryBlobRef:
        """
        Writes a blob to the registry. The digest will be calculated
        automatically while uploading, ignoring `ref.ref`. A copy of
        `ref` with `ref.ref` set to the calculated digest will be returned.

        progress_callback is an optional coroutine function that if given will
        be called periodically as data is written with the number of bytes
        written. It is guaranteed that the last call for a successful write
        will be the total number of bytes written.
        """
        # Perform the blob upload flow, POST -> PATCH -> PUT
        registry = ref.registry or self.default_registry
        async with await self._request(
            "POST",
            registry,
            ref.upload_url(),
        ) as response:
            if response.status != 202:
                raise RegistryException(
                    "Unexpected response attempting to start blob copy"
                )
            upload_location = response.headers["Location"]

        if progress_callback is not None:
            await progress_callback(0)

        hsh = hashlib.sha256()
        offset = 0
        async for chunk in async_generator_buffer(data, 4):
            LOGGER.debug("Writing chunk %d - %d", offset, offset + len(chunk) - 1)

            headers = {
                "Content-Length": str(len(chunk)),
                "Content-Range": f"{offset}-{offset+len(chunk)-1}",
                "Content-Type": "application/octet-stream",
            }
            hsh.update(chunk)
            async with await self._request(
                "PATCH",
                registry,
                upload_location,
                data=chunk,
                headers=headers,
            ) as response:
                if response.status // 100 != 2:
                    raise RegistryException("Unexpected response writing blob data")
                upload_location = urllib.parse.urljoin(
                    upload_location,
                    response.headers["Location"],
                )
            offset += len(chunk)
            if progress_callback is not None:
                await progress_callback(offset)

        digest = "sha256:" + hsh.hexdigest()

        try:
            upload_url = urllib.parse.urlparse(upload_location)
            query_data = urllib.parse.parse_qsl(upload_url.query)
        except ValueError as exc:
            raise RegistryException(
                "Unexpected upload URL format from registry"
            ) from exc
        query_data.append(("digest", digest))

        async with await self._request(
            "PUT",
            registry,
            upload_url._replace(query=urllib.parse.urlencode(query_data)).geturl(),
        ) as response:
            if response.status // 100 != 2:
                raise RegistryException("Unexpected response ending blob copy")

        return ref.copy(update={"ref": digest})

    async def registry_repos(self, registry: Optional[Registry]) -> List[str]:
        """
        Return a list of all repos for the given registry. It is up to the
        registry implementation to determine what if any repo names will
        be returned.
        """
        async with await self._request(
            "GET",
            registry or self.default_registry,
            "/v2/_catalog",
        ) as response:
            try:
                return (await response.json())["repositories"]
            except ValueError as exc:
                raise RegistryException("Unexpected response getting repos") from exc

    async def registry_repo_tags(
        self, registry: Optional[Registry], repo: List[str]
    ) -> List[str]:
        """
        Return a list of all tags for the given repo name.
        """
        async with await self._request(
            "GET",
            registry or self.default_registry,
            f"/v2/{'/'.join(repo)}/tags/list",
        ) as response:
            try:
                return (await response.json())["tags"]
            except ValueError as exc:
                raise RegistryException(
                    "Unexpected response getting repo tags"
                ) from exc

    async def copy_refs(
        self,
        src: RegistryBlobRef,
        dst: RegistryBlobRef,
        *,
        layer_progress: Optional[
            Callable[[RegistryBlobRef, RegistryBlobRef, int, int], Awaitable[None]]
        ] = None,
    ) -> bool:
        """
        Copy the blob src to dst. Returns True if any data was copied and
        False if the content already existed.

        layer_progress can be provided as a coroutine function take takes arguments
        (src, dst, total_bytes, written_bytes). It will be called periodically when
        copying blob data. It is guaranteed that for a successful copy the last call
        will have total_bytes == written_bytes.
        """
        if src.OBJECT_TYPE != dst.OBJECT_TYPE:
            raise ValueError("Cannot copy ref to different object type")
        if dst.is_digest_ref():
            if src.ref != dst.ref:
                raise ValueError(
                    "Cannot copy to a content address that does not match the source"
                )

        if src == dst:
            LOGGER.info("skipping copy of identical refs")
            return False

        # Check if ref already exists
        if src.is_digest_ref():
            if await self.ref_lookup(dst) is not None:
                LOGGER.info("Skipping copy %s -> %s - already exists", src, dst)
                return False

        if isinstance(src, RegistryManifestRef):
            manifest = await self.manifest_download(src)

            await asyncio.gather(
                *(
                    self.copy_refs(
                        RegistryManifestRef(
                            registry=src.registry, repo=src.repo, ref=digest
                        ),
                        RegistryManifestRef(
                            registry=dst.registry, repo=dst.repo, ref=digest
                        ),
                        layer_progress=layer_progress,
                    )
                    for digest in manifest.get_manifest_dependencies()
                ),
                *(
                    self.copy_refs(
                        RegistryBlobRef(
                            registry=src.registry, repo=src.repo, ref=digest
                        ),
                        RegistryBlobRef(
                            registry=dst.registry, repo=dst.repo, ref=digest
                        ),
                        layer_progress=layer_progress,
                    )
                    for digest in manifest.get_blob_dependencies()
                ),
            )
            assert isinstance(dst, RegistryManifestRef)
            await self.manifest_write(dst, manifest)

            LOGGER.info("Copied manifest %s -> %s", src, dst)
            return True

        # Attempt mount if blobs come from the same registry.
        if src.registry == dst.registry:
            query_str = urllib.parse.urlencode(
                {
                    "from": "/".join(src.repo),
                    "mount": dst.ref,
                }
            )
            async with await self._request(
                "POST",
                dst.registry or self.default_registry,
                f"v2/{'/'.join(dst.repo)}/blobs/uploads/?{query_str}",
            ) as response:
                if response.status == 201:
                    LOGGER.info(
                        "Mounted blob %s from %s",
                        dst.ref,
                        "/".join(src.repo),
                    )
                    return True
                LOGGER.warning(
                    "mount failed with status %s, trying copy",
                    response.status,
                )

        async def write_callback(
            bytes_written: int,
        ) -> None:  # pylint: disable=function-redefined
            if layer_progress is not None:
                await layer_progress(src, dst, desc.size, bytes_written)

        desc, content_stream = await self.ref_content_stream(src)
        await self.blob_write(dst, content_stream, progress_callback=write_callback)
        LOGGER.info("Copied blob %s -> %s", src, dst)
        return True



#!/usr/bin/env python3
"""
Script entrypoint for copying images between registries.
"""

import argparse
import asyncio
import json
import logging
import re
import ssl
import sys
from typing import Any, Callable, Dict, Optional, Tuple

import pydantic

try:
    from tqdm import tqdm  # type: ignore
    from tqdm.contrib.logging import logging_redirect_tqdm  # type: ignore
except ImportError:
    tqdm = None
    logging_redirect_tqdm = None

from .auth import CredentialStore, DockerCredentialStore, default_credential_store
from .client import AsyncRegistryClient
from .models import RegistryBlobRef, RegistryManifestRef
from .parsing import parse_image_name


def parse_args():
    """
    Setup and parse CLI arguments.
    """
    parser = argparse.ArgumentParser(
        description="Copy/inspect registry images",
    )
    parser.add_argument("src", help="Source registry image")
    parser.add_argument("dst", nargs="?", help="Dest registry image")
    parser.add_argument(
        "--blob",
        required=False,
        const=True,
        action="store_const",
        default=False,
        help="Copy/inspect a blob instead of a manifest",
    )
    parser.add_argument(
        "--tag-pattern",
        action="append",
        help="Copy/inspect all tags matching regex. Not compatible with --blob",
    )
    parser.add_argument(
        "--descriptor",
        required=False,
        const=True,
        action="store_const",
        default=False,
        help="Print a descriptor of the objects instead of the objects themselves",
    )
    parser.add_argument(
        "--auth-config",
        required=False,
        default=None,
        help="Path to Docker credential config file",
    )
    parser.add_argument(
        "--insecure",
        required=False,
        const=True,
        action="store_const",
        default=False,
        help="Disable server certificate verification",
    )
    parser.add_argument(
        "--cafile",
        required=False,
        default=None,
        help="SSL context CA file",
    )
    parser.add_argument(
        "--capath",
        required=False,
        default=None,
        help="SSL context CA directory",
    )
    parser.add_argument(
        "--verbose",
        "-v",
        action="count",
        default=0,
    )
    return parser.parse_args()


def setup_logging(verbose: int) -> None:
    """
    Setup logging configuration for CLI
    """
    log_level = logging.WARN
    if verbose > 1:
        log_level = logging.DEBUG
    elif verbose:
        log_level = logging.INFO
    logging.basicConfig(
        level=log_level,
        format="%(message)s",
    )


def _convert_to_blob_ref(ref: RegistryManifestRef) -> RegistryBlobRef:
    """
    Convert a manifest ref into a blob ref of the same repo/ref.
    """
    if not ref.is_digest_ref():
        sys.stderr.write("Can only copy/inspect blobs by digest")
        sys.exit(1)
    return RegistryBlobRef(
        registry=ref.registry,
        repo=ref.repo,
        ref=ref.ref,
    )


async def main() -> int:
    """
    CLI entrypoint that copies an image between two registries.
    """
    args = parse_args()
    setup_logging(args.verbose)

    if logging_redirect_tqdm is not None:
        with logging_redirect_tqdm():
            return await _main(args)

    return await _main(args)


async def _main(args) -> int:
    """
    Helper CLI entrypoint after logging setup and arguments parsed.
    """
    creds: CredentialStore
    if args.auth_config:
        with open(args.auth_config, "r", encoding="utf-8") as fauth:
            creds = DockerCredentialStore(json.load(fauth))
    else:
        creds = default_credential_store()  # pylint: disable=redefined-variable-type

    ssl_ctx = ssl.create_default_context(
        cafile=args.cafile,
        capath=args.capath,
    )
    if args.insecure:
        ssl_ctx.check_hostname = False
        ssl_ctx.verify_mode = ssl.CERT_NONE

    src_ref = parse_image_name(args.src)

    progress_map: Dict[int, Tuple[Any, int]] = {}

    async def _progress(src, dst, bytes_total, bytes_written):
        # pylint: disable=unused-argument
        if tqdm is None:
            print(
                f"Copying {dst} {bytes_written}/{bytes_total} {100*bytes_written/bytes_total:.1f}%"
            )
        else:
            t, bytes_last = progress_map.get(id(dst), (None, 0))
            if t is None:
                # pylint: disable=unnecessary-dunder-call
                t = tqdm(
                    desc=dst.ref[7:15],
                    total=bytes_total,
                    unit="B",
                    unit_scale=True,
                    unit_divisor=1024,
                ).__enter__()
            t.update(bytes_written - bytes_last)
            progress_map[id(dst)] = (t, bytes_written)

    async with AsyncRegistryClient(creds=creds, ssl_context=ssl_ctx) as client:

        def _convert_object(obj: Optional[pydantic.BaseModel]):
            if obj is None:
                return None
            return obj.dict(
                exclude_unset=True,
                by_alias=True,
            )

        get_object: Callable = (
            client.ref_lookup if args.descriptor else client.manifest_download
        )

        if not args.dst:
            if args.blob:
                if args.descriptor:
                    result = _convert_object(
                        await client.ref_lookup(_convert_to_blob_ref(src_ref))
                    )
                else:
                    _, content_stream = await client.ref_content_stream(
                        _convert_to_blob_ref(src_ref)
                    )
                    async for chunk in content_stream:
                        sys.stdout.buffer.write(chunk)
                    return 0

            elif args.tag_pattern:
                result = {}
                for tag in await client.registry_repo_tags(
                    src_ref.registry, src_ref.repo
                ):
                    if not any(re.match(pat, tag) for pat in args.tag_pattern):
                        continue
                    the_ref = src_ref.copy(update={"ref": tag})
                    result[tag] = _convert_object(await get_object(the_ref))
            else:
                result = _convert_object(await get_object(src_ref))
            json.dump(result, sys.stdout, indent=2)
            sys.stdout.write("\n")
            return 0

        dst_ref = parse_image_name(args.dst)
        if args.blob:
            await client.copy_refs(
                _convert_to_blob_ref(src_ref),
                _convert_to_blob_ref(dst_ref),
                layer_progress=_progress,
            )
        elif args.tag_pattern:
            for tag in await client.registry_repo_tags(src_ref.registry, src_ref.repo):
                if not any(re.match(pat, tag) for pat in args.tag_pattern):
                    continue
                print(f"Copying {src_ref} to {dst_ref}")
                await client.copy_refs(
                    src_ref.copy(update={"ref": tag}),
                    dst_ref.copy(update={"ref": tag}),
                    layer_progress=_progress,
                )
        else:
            await client.copy_refs(src_ref, dst_ref, layer_progress=_progress)

    return 0


def sync_main() -> int:
    """Synchronous entry point"""
    return asyncio.run(main())



from typing import List, Optional

from .models import Registry, RegistryManifestRef


def parse_image_name(name: str) -> RegistryManifestRef:
    """
    Extract out the registry host, image repo, and tag from an image string.

    urllib.parse does not work appropriately for this task.
    """
    # Extract protocol if present. If no protocol is present it will be
    # set to a default value depending on the hostname.
    prot = None
    prot_start = name.find("://")
    if prot_start != -1:
        prot = name[0:prot_start].lower()
        name = name[prot_start + 3 :]

    # Extract the registry host if the first token looks like a hostname. If a
    # protocol scheme was present always treat the first token as the host.
    registry: Optional[Registry]
    reg_part, *slash_parts = name.split("/")
    if prot is not None or (
        (":" in reg_part or "." in reg_part or reg_part == "localhost") and slash_parts
    ):
        reg_part = reg_part.lower()
        host_alias = (prot + "://" if prot else "") + reg_part

        # Extract port
        port = None
        port_start = reg_part.rfind(":")
        if port_start != -1:
            port = int(reg_part[port_start + 1 :])
            reg_part = reg_part[0:port_start]

        # If no prot specified fill in default
        if prot is None:
            if port == 80:
                prot = "http"
            elif port == 443:
                prot = "https"
            else:
                prot = "http" if reg_part in ("localhost", "127.0.0.1") else "https"

        if prot not in ("http", "https"):
            raise ValueError("unknown registry protocol")

        if port is None:
            port = 443 if prot == "https" else 80

        if reg_part == "docker.io":
            reg_part = "registry-1.docker.io"

        registry = Registry(
            host=reg_part,
            port=port,
            prot=prot,
            host_alias=host_alias if host_alias != reg_part else None,
        )
    else:
        registry = None
        slash_parts.insert(0, reg_part)

        # Bare images with no slash are prefixed with library. e.g.
        # ubuntu becomes docker.io/library/ubuntu:latest.
        if len(slash_parts) == 1:
            slash_parts.insert(0, "library")

    if not slash_parts:
        raise ValueError("No repo name")

    # Extract out the tag specifier.
    tag = "latest"
    try:
        tag_start = next(i for i, ch in enumerate(slash_parts[-1]) if ch in ":@")
    except StopIteration:
        tag_start = -1

    if tag_start != -1:
        tag = slash_parts[-1][tag_start + 1 :]
        slash_parts[-1] = slash_parts[-1][0:tag_start]

    return RegistryManifestRef(registry=registry, repo=slash_parts, ref=tag)


def split_quote(s: str, dels: str, quotes: str = '"', escape: str = "\\") -> List[str]:
    """
    Split s by any character present in dels. However treat anything
    surrounded by a character in quotes as a literal. Additionally
    any character preceeded by escape is treated as a literal.

    Returns a list of split tokens with the split delimeter between each token.
    The length of the result will always be odd with the even indexed elements
    being the split data and the odd indexed elements being the delimeters
    between the even elements.

    _split_quote('a="b,c",d=f', '=,') => ['a', '=', 'b,c', ',', 'd', '=', 'f']
    """
    part: List[str] = []
    result: List[str] = []

    quote = None
    for ch in s:
        if part and part[-1] == escape:
            part[-1] = ch
        elif quote and ch == quote:
            quote = None
        elif quote:
            part.append(ch)
        elif ch in dels:
            result.append("".join(part))
            result.append(ch)
            part.clear()
        elif ch in quotes:
            quote = ch
        else:
            part.append(ch)
    result.append("".join(part))

    return result
